import tempfile
from itertools import product
from pathlib import Path
from typing import Optional

import pytest
import torch

import swyft.inference.marginalratioestimator as mre
import swyft.networks.classifier as classifier
from swyft.types import MarginalIndex
from swyft.utils.marginals import tupleize_marginal_indices


class TestSplitLengthByPercentage:
    def test_sum_to_1(self):
        """The percentage vector should sum to one."""
        with pytest.raises(AssertionError):
            mre.split_length_by_percentage(10, [0.2, 0.3])

    def test_extra_in_first_position(self):
        """The rounding error goes into the first index."""
        out = mre.split_length_by_percentage(10, [0.61, 0.39])
        assert out == [7, 3]


class TestNTrainNValid:
    dataset_lens = [9, 25, 100]
    validation_amounts = [1, 2, 0.23, 0.5]

    @pytest.mark.parametrize(
        "validation_amount, len_dataset", product(validation_amounts, dataset_lens)
    )
    def test_nvalid_divisible_by_2(self, validation_amount, len_dataset):
        """nvalid should be divisible by 2."""
        _, nvalid = mre.get_ntrain_nvalid(validation_amount, len_dataset)
        assert nvalid % 2 == 0

    def test_wrong_type(self):
        """A non-int or non-float for validation_amount should raise a TypeError."""
        with pytest.raises(TypeError):
            mre.get_ntrain_nvalid(complex(5), 100)

    @pytest.mark.parametrize(
        "validation_amount, len_dataset", product(validation_amounts, dataset_lens)
    )
    def test_nvalid_divisible_by_2(self, validation_amount, len_dataset):
        """nvalid should be divisible by 2."""
        _, nvalid = mre.get_ntrain_nvalid(validation_amount, len_dataset)
        assert nvalid % 2 == 0


def test_double_observation():
    b, o = 2, 3
    a = torch.arange(b * o).reshape(b, o)
    truth = torch.Tensor([[0, 1, 2], [0, 1, 2], [3, 4, 5], [3, 4, 5]])
    assert torch.all(mre.double_observation(a) == truth)


class TestDoubleParameters:
    def test_assertion(self):
        with pytest.raises(AssertionError):
            mre.double_parameters(torch.rand(5, 2))

    def test_shape(self):
        with pytest.raises(ValueError):
            mre.double_parameters(torch.rand(6, 2, 2))

    def test_correcness(self):
        b, p = 2, 3
        a = torch.arange(b * p).reshape(b, p)
        truth = torch.Tensor([[0, 1, 2], [3, 4, 5], [0, 1, 2], [3, 4, 5]])
        assert torch.all(mre.double_parameters(a) == truth)


class TestSaveLoadMarginalRatioEstimator:
    @classmethod
    def setup_class(cls):
        cls.directory = tempfile.TemporaryDirectory()

    @classmethod
    def teardown_class(cls):
        cls.directory.cleanup()

    def test_save_load_marginal_ratio_estimator(self):
        # Preliminary
        observation_key = "x"
        n_parameters = 2
        marginal_indices = list(range(n_parameters))
        device = "cpu"

        network = classifier.get_marginal_classifier(
            observation_key=observation_key,
            marginal_indices=marginal_indices,
            observation_shapes={observation_key: (10,)},
            n_parameters=n_parameters,
            hidden_features=32,
            num_blocks=2,
        )

        marginal_ratio_estimator = mre.MarginalRatioEstimator(
            marginal_indices=marginal_indices,
            network=network,
            device=device,
        )

        # Saving
        path = Path(self.directory.name) / f"marginal_ratio_estimator"
        marginal_ratio_estimator.save(path)

        # Loading
        marginal_ratio_estimator_loaded = mre.MarginalRatioEstimator.load(
            network=network,
            device=device,
            filename=path,
        )

        # Testing each loaded piece
        for weights, weights_loaded in zip(
            marginal_ratio_estimator.network.parameters(),
            marginal_ratio_estimator_loaded.network.parameters(),
        ):
            assert torch.all(weights == weights_loaded)

        for buffers, buffers_loaded in zip(
            marginal_ratio_estimator.network.buffers(),
            marginal_ratio_estimator_loaded.network.buffers(),
        ):
            assert torch.all(buffers == buffers_loaded)

        assert marginal_ratio_estimator.device == marginal_ratio_estimator_loaded.device

        # TODO: need to add a test for scheduler and optimizer loading


class TestMarginalRatioEstimator:
    @classmethod
    def setup_class(cls):
        cls.observation_key = "x"
        cls.observation_shapes = {cls.observation_key: (10,)}
        cls.n_parameters = 2
        cls.device = "cpu"

    @classmethod
    def get_marginal_ratio_estimator(cls, marginal_indices):
        network = classifier.get_marginal_classifier(
            observation_key=cls.observation_key,
            marginal_indices=marginal_indices,
            observation_shapes=cls.observation_shapes,
            n_parameters=cls.n_parameters,
            hidden_features=16,
            num_blocks=1,
        )
        marginal_ratio_estimator = mre.MarginalRatioEstimator(
            marginal_indices=marginal_indices,
            network=network,
            device=cls.device,
        )
        return marginal_ratio_estimator

    @pytest.mark.skip
    def test_train(self):
        raise NotImplementedError("Need to test this function.")

    @pytest.mark.parametrize(
        "marginal_indices, batch_size",
        product(
            [[0, 1], [(0, 1)]],  # With these, n_parameters >= 2
            [None, 10],
        ),
    )
    def test_log_ratio_shape(
        self, marginal_indices: MarginalIndex, batch_size: Optional[int]
    ):
        """The log_ratio function should return as many weights as there were n_batches of parameters provided."""
        n_batch = 100
        marginal_indices = tupleize_marginal_indices(marginal_indices)
        marginal_ratio_estimator = self.get_marginal_ratio_estimator(marginal_indices)
        fabricated_observation = {
            key: torch.rand(*shape) for key, shape in self.observation_shapes.items()
        }
        fabricated_v = torch.randn(n_batch, self.n_parameters)
        log_ratio = marginal_ratio_estimator.log_ratio(
            observation=fabricated_observation,
            v=fabricated_v,
            batch_size=batch_size,
        )
        assert set(log_ratio.keys()) == set(marginal_indices)
        for _, value in log_ratio.items():
            assert value.shape == (n_batch,)


if __name__ == "__main__":
    pass



# pylint: disable=no-member, not-callable
import logging
import os
import time
from pathlib import Path
from typing import Mapping, Optional, Sequence, Tuple, Union

import fasteners
import numcodecs
import numpy as np
import zarr

import swyft
from swyft.store.simulator import SimulationStatus, Simulator
from swyft.types import Array, ObsShapeType, ParameterNamesType, PathType
from swyft.utils import is_empty

log = logging.getLogger(__name__)


class Filesystem:
    metadata = "metadata"
    log_lambdas = "metadata/log_lambdas"
    samples = "samples"
    sims = "samples/sims"
    v = "samples/v"
    log_w = "samples/log_w"
    simulation_status = "samples/simulation_status"


class Store:
    """Store of sample parameters and simulation outputs.

    Based on Zarr, it should be instantiated via its methods `memory_store`,
    `directory_store` or `load`.

    Args:
        zarr_store: Zarr store object.
        simulator: simulator object.
        sync_path: if specified, it will enable synchronization using file locks (files will be
            stored in the given path). Must be accessible to all processes working on the store
            and the underlying filesystem must support file locking.
        chunksize: the parameters and simulation output will be stored as arrays with the
            specified chunk size along the sample dimension (a single chunk will be used for the
            other dimensions).
        pickle_protocol: pickle protocol number used for storing intensity functions.
        from_scratch: if False, load the sample store from the Zarr store provided.
    """

    _filesystem = Filesystem

    def __init__(
        self,
        zarr_store: Union[zarr.MemoryStore, zarr.DirectoryStore],
        simulator: Optional[Simulator] = None,
        sync_path: Optional[PathType] = None,
        chunksize: int = 1,
        pickle_protocol: int = 4,
        from_scratch: bool = True,
    ) -> None:
        self._zarr_store = zarr_store
        self._simulator = simulator
        self._pickle_protocol = pickle_protocol  # TODO: to be deprecated, we will default to 4, which is supported since python 3.4

        synchronizer = zarr.ProcessSynchronizer(sync_path) if sync_path else None
        self._root = zarr.group(
            store=self._zarr_store, synchronizer=synchronizer, overwrite=from_scratch
        )

        if not from_scratch:
            if not {"samples", "metadata"} == self._root.keys():
                raise KeyError(
                    "Invalid Zarr store. It should have keys: ['samples', 'metadata']."
                )
            print("Loading existing store.")
            self._update()
        else:
            print("Creating new store.")
            if simulator is None:
                raise ValueError("A simulator is required to setup a new store.")
            self._setup_new_zarr_store(
                simulator.parameter_names,
                simulator.sim_shapes,
                self._root,
                chunksize=chunksize,
                sim_dtype=simulator.sim_dtype,
            )
            log.debug("  sim_shapes = %s" % str(simulator.sim_shapes))

        # a second layer of synchronization is required to grow the store
        self._lock = None
        if sync_path is not None:
            self._setup_lock(sync_path)

    def add(
        self, N: int, prior: "swyft.Prior", bound: Optional["swyft.Bound"] = None
    ) -> None:
        """Adds points to the store.

        Args:
            N: Number of samples
            prior: Prior
            bound: Bound object for prior truncation

        .. warning::
            Calling this method will alter the content of the store by adding
            additional points. Currently this cannot be reverted, so use with
            care when applying it to the DirectoryStore.
        """
        pdf = swyft.PriorTruncator(prior, bound)

        # Lock store while adding new points
        self.lock()
        self._update()

        # Generate new points
        v_prop = pdf.sample(np.random.poisson(N))
        log_lambda_target = pdf.log_prob(v_prop) + np.log(N)
        log_lambda_store = self.log_lambda(v_prop)
        log_w = np.log(np.random.rand(len(v_prop))) + log_lambda_target
        accept_new = log_w > log_lambda_store
        v_new = v_prop[accept_new]
        log_w_new = log_w[accept_new]

        # Anything new?
        if sum(accept_new) > 0:
            # Add new entries to store
            self._append_new_points(v_new, log_w_new)
            print("Store: Adding %i new samples to simulator store." % sum(accept_new))
            # Update intensity function
            self.log_lambdas.resize(len(self.log_lambdas) + 1)
            self.log_lambdas[-1] = dict(pdf=pdf.state_dict(), N=N)

        log.debug(f"  total size of simulator store {len(self)}.")

        # Points added, unlock store
        self.unlock()
        self._update()

    #    @property
    #    def zarr_store(self):
    #        """Return ZarrStore object."""
    #        return self._zarr_store

    def _setup_lock(self, sync_path: PathType) -> None:
        """Setup lock for concurrent access from multiple processes."""
        path = os.path.join(sync_path, "cache.lock")
        self._lock = fasteners.InterProcessLock(path)

    def lock(self) -> None:
        """Lock store for the current process."""
        if self._lock is not None:
            log.debug("Store locked")
            self._lock.acquire(blocking=True)

    def unlock(self) -> None:
        """Unlock store so that other processes can access it."""
        if self._lock is not None:
            self._lock.release()
            log.debug("Store unlocked")

    def _setup_new_zarr_store(
        self,
        parameter_names: ParameterNamesType,
        sim_shapes: ObsShapeType,
        root: zarr.Group,
        chunksize: int = 1,
        sim_dtype: str = "f8",
    ) -> None:  # Adding observational shapes to store
        # Parameters
        n_parameters = len(parameter_names)
        v = root.zeros(
            self._filesystem.v,
            shape=(0, n_parameters),
            chunks=(chunksize, n_parameters),
            dtype="f8",
        )
        v.attrs["parameter_names"] = parameter_names

        # Simulations
        sims = root.create_group(self._filesystem.sims)
        for name, shape in sim_shapes.items():
            sims.zeros(
                name, shape=(0, *shape), chunks=(chunksize, *shape), dtype=sim_dtype
            )

        # Random intensity weights
        root.zeros(self._filesystem.log_w, shape=(0,), chunks=(chunksize,), dtype="f8")

        # Pickled Intensity (prior * N) objects
        root.create(
            self._filesystem.log_lambdas,
            shape=(0,),
            dtype=object,
            object_codec=numcodecs.Pickle(protocol=self._pickle_protocol),
        )

        # Simulation status code
        root.zeros(
            self._filesystem.simulation_status,
            shape=(0,),
            chunks=(chunksize,),
            dtype="int",
        )

        self._update()

    def _update(self) -> None:
        self.sims = self._root[self._filesystem.sims]
        self.v = self._root[self._filesystem.v]
        self.log_w = self._root[self._filesystem.log_w]
        self.log_lambdas = self._root[self._filesystem.log_lambdas]
        self.sim_status = self._root[self._filesystem.simulation_status]
        self.parameter_names = self._root[self._filesystem.v].attrs["parameter_names"]

    def __len__(self) -> int:
        """Returns number of samples in the store."""
        self._update()
        return len(self.v)

    def __getitem__(self, i: int) -> Tuple[Mapping[str, np.ndarray], np.ndarray]:
        """Returns data store entry with index :math:`i`."""
        self._update()
        sim = {}
        for key, value in self.sims.items():
            sim[key] = value[i]
        par = self.v[i]
        return (sim, par)

    def _append_new_points(self, v: Array, log_w: Array) -> None:
        """Append z to zarr_store content and generate new slots for x."""
        self._update()
        n = len(v)
        for key, value in self.sims.items():
            shape = list(value.shape)
            shape[0] += n
            value.resize(*shape)

        self._root[self._filesystem.v].append(v)
        self._root[self._filesystem.log_w].append(log_w)
        m = np.full(n, SimulationStatus.PENDING, dtype="int")
        self.sim_status.append(m)

    def log_lambda(self, z: np.ndarray) -> np.ndarray:
        """Intensity function of the store.

        Args:
            z: Array with the sample parameters. Should have shape (num. samples,
                num. parameters per sample).

        Returns:
            Array with the sample intensities.
        """
        self._update()
        d = -np.inf * np.ones_like(z[:, 0])
        if len(self.log_lambdas) == 0:
            return d
        for i in range(len(self.log_lambdas)):
            pdf = swyft.PriorTruncator.from_state_dict(self.log_lambdas[i]["pdf"])
            N = self.log_lambdas[i]["N"]
            r = pdf.log_prob(z) + np.log(N)
            d = np.where(r > d, r, d)
        return d

    def coverage(
        self, N: int, prior: "swyft.Prior", bound: Optional["swyft.Bound"] = None
    ) -> float:
        """Returns fraction of already stored data points.

        Args:
            N: Number of samples
            prior: Prior
            bound: Bound object for prior truncation

        Returns:
            Fraction of samples that is already covered by content of the store.

        .. note::
            A coverage of zero means that all points need to be newly
            simulated. A coverage of 1.0 means that all points are already
            available for this (truncated) prior.

        .. warning::
            Results are Monte Carlo estimated and subject to sampling noise.
        """
        pdf = swyft.PriorTruncator(prior, bound)
        Nsamples = max(N, 1000)  # At least 1000 test samples
        self._update()

        # Generate new points
        v_prop = pdf.sample(np.random.poisson(Nsamples))
        log_lambda_target = pdf.log_prob(v_prop) + np.log(N)
        log_lambda_store = self.log_lambda(v_prop)
        frac = np.where(
            log_lambda_target > log_lambda_store,
            np.exp(-log_lambda_target + log_lambda_store),
            1.0,
        ).mean()
        return frac

    def sample(
        self,
        N: int,
        prior: "swyft.Prior",
        bound: Optional["swyft.Bound"] = None,
        check_coverage: bool = True,
        add: bool = False,
    ) -> np.ndarray:
        """Return samples from store.

        Args:
            N: Number of samples
            prior: Prior
            bound: Bound object for prior truncation
            check_coverage: Check whether requested points are contained in the store.
            add: If necessary, add requested points to the store.

        Returns:
            Indices: Index list pointing to the relevant store entries.
        """
        if add:
            if self.coverage(N, prior, bound=bound) < 1:
                self.add(N, prior, bound=bound)
        if check_coverage:
            if self.coverage(N, prior, bound=bound) < 1.0:
                raise RuntimeError(
                    "Store does not contain enough samples for your requested intensity function `N * prior`."
                )
        pdf = swyft.PriorTruncator(prior, bound)

        self._update()

        # Select points from cache
        v_store = self.v[:]
        log_w_store = self.log_w[:]
        log_lambda_target = pdf.log_prob(v_store) + np.log(N)
        accept_stored = log_w_store <= log_lambda_target
        indices = np.array(range(len(accept_stored)))[accept_stored]

        return indices

    def _get_indices_to_simulate(
        self, indices: Optional[Sequence[int]] = None
    ) -> np.ndarray:
        """
        Determine which samples need to be simulated.

        Args:
            indices: array with the indices of the samples to
            consider. If None, consider all samples.

        Returns:
            array with the sample indices
        """
        status = self.get_simulation_status(indices)
        require_simulation = status == SimulationStatus.PENDING
        idx = np.flatnonzero(require_simulation)
        return indices[idx] if indices is not None else idx

    def _set_simulation_status(
        self, indices: Sequence[int], status: SimulationStatus
    ) -> None:
        """
        Flag the specified samples with the simulation status.

        Args:
            indices: array with the indices of the samples to flag
            status: new status for the samples
        """
        assert status in list(SimulationStatus), f"Unknown status {status}"
        current_status = self.sim_status.oindex[indices]
        if np.any(current_status == status):
            log.warning(
                f"Changing simulation status to {status}, but some simulations have already status {status}"
            )
        self.sim_status.oindex[indices] = status

    def get_simulation_status(
        self, indices: Optional[Sequence[int]] = None
    ) -> np.ndarray:
        """Determine the status of sample simulations.

        Args:
            indices: List of indices. If None, check the status of all
                samples

        Returns:
            list of simulation statuses
        """
        self._update()
        return (
            self.sim_status.oindex[indices]
            if indices is not None
            else self.sim_status[:]
        )

    def requires_sim(self, indices: Optional[Sequence[int]] = None) -> bool:
        """Check whether there are parameters which require simulation.

        Args:
            indices: List of indices. If None, check all samples.

        Returns:
            True if one or more samples require simulations, False otherwise.
        """
        self._update()
        return self._get_indices_to_simulate(indices).size > 0

    def _get_indices_failed_simulations(self) -> np.ndarray:
        self._update()
        return np.flatnonzero(self.sim_status == SimulationStatus.FAILED)

    @property
    def any_failed(self) -> bool:
        """Check whether there are parameters which currently lead to a failed simulation."""
        self._update()
        return self._get_indices_failed_simulations().size > 0

    def _add_sim(self, i: int, x: Mapping[str, Array]) -> None:
        for k, v in x.items():
            self.sims[k][i] = v
        self._set_simulation_status(i, SimulationStatus.FINISHED)

    def _failed_sim(self, i: int) -> None:
        self._update()
        self._set_simulation_status(i, SimulationStatus.FAILED)

    def set_simulator(self, simulator: "swyft.Simulator") -> None:
        """(Re)set simulator.

        Args:
            simulator: Simulator.
        """
        if self._simulator is not None:
            log.warning("Simulator already set!  Overwriting.")
        self._simulator = simulator

    def simulate(
        self,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = None,
        wait_for_results: Optional[bool] = True,
    ) -> None:
        """Run simulator on parameter store with missing corresponding simulations.

        Args:
            indices: list of sample indices for which a simulation is required
            batch_size: simulations will be submitted in batches of the specified size
            wait_for_results: if True, return only when all simulations are done
        """
        if self._simulator is None:
            log.warning("No simulator specified.  No simulations will run.")
            return

        self.lock()
        self._update()
        idx = self._get_indices_to_simulate(indices)
        self._set_simulation_status(idx, SimulationStatus.RUNNING)
        self.unlock()

        # Run simulations and collect status
        if len(idx) == 0:
            log.debug("No simulations required.")
        else:
            # For the MemoryStore, we need to collect results in memory
            collect_in_memory = (
                True if isinstance(self._zarr_store, zarr.MemoryStore) else False
            )

            if collect_in_memory and not wait_for_results:
                logging.warning(
                    "Asynchronous collection of results is not implemented with the MemoryStore"
                )

            self._simulator._run(
                v=self.v,
                sims={k: v.oindex for k, v in self.sims.items()},
                sim_status=self.sim_status.oindex,
                indices=idx,
                collect_in_memory=collect_in_memory,
                batch_size=batch_size,
            )

        if wait_for_results:
            self.wait_for_simulations(indices)

    def wait_for_simulations(self, indices: Sequence[int]) -> None:
        """Wait for a set of sample simulations to be finished.

        Args:
            indices: list of sample indices
        """
        done = False
        while not done:
            time.sleep(1)
            status = self.get_simulation_status(indices)
            done = np.isin(status, [SimulationStatus.FINISHED, SimulationStatus.FAILED])
            done = np.all(done)

    @classmethod
    def directory_store(
        cls,
        path: PathType,
        simulator: Optional[Simulator] = None,
        sync_path: Optional[PathType] = None,
        overwrite: bool = False,
    ) -> "Store":
        """Instantiate a new Store based on a Zarr DirectoryStore.

        Args:
            path: path to storage directory
            simulator: simulator object
            sync_path: path for synchronization via file locks (files will be stored in the given path).
                It must differ from path, it must be accessible to all processes working on the store,
                and the underlying filesystem must support file locking.
            overwrite: if True, and a store already exists at the specified path, overwrite it.

        Returns:
            Store based on a Zarr DirectoryStore

        Example:
            >>> store = swyft.Store.directory_store(PATH_TO_STORE)
        """
        if not Path(path).exists() or overwrite:
            zarr_store = zarr.DirectoryStore(path)
            sync_path = sync_path or os.path.splitext(path)[0] + ".sync"
            return cls(
                zarr_store=zarr_store,
                simulator=simulator,
                sync_path=sync_path,
                from_scratch=True,
            )
        else:
            raise FileExistsError(
                f"Path {path} exists - set overwrite=True to initialize a new store there."
            )

    @classmethod
    def memory_store(cls, simulator: Simulator) -> "Store":
        """Instantiate a new Store based on a Zarr MemoryStore.

        Args:
            simulator: simulator object

        Returns:
            Store based on a Zarr MemoryStore

        Note:
            The store returned is in general expected to be faster than an equivalent
            store based on the Zarr DirectoryStore, and thus useful for quick
            explorations, or for loading data into memory before training.

        Example:
            >>> store = swyft.Store.memory_store(simulator)
        """
        zarr_store = zarr.MemoryStore()
        return cls(zarr_store=zarr_store, simulator=simulator, from_scratch=True)

    def save(self, path: PathType) -> None:
        """Save the Store to disk using a Zarr DirectoryStore.

        Args:
            path: path where to create the Zarr root directory
        """
        if isinstance(
            self._zarr_store, zarr.DirectoryStore
        ) and self._zarr_store.path == os.path.abspath(path):
            return
        path = Path(path)
        if path.exists() and not path.is_dir():
            raise NotADirectoryError(f"{path} should be a directory")
        elif path.exists() and not is_empty(path):
            raise FileExistsError(f"{path} is not empty")
        else:
            path.mkdir(parents=True, exist_ok=True)
            zarr_store = zarr.DirectoryStore(path)
            zarr.convenience.copy_store(source=self._zarr_store, dest=zarr_store)

    @classmethod
    def load(
        cls,
        path: PathType,
        simulator: Optional[Simulator] = None,
        sync_path: Optional[PathType] = None,
    ) -> "Store":
        """Open an existing sample store using a Zarr DirectoryStore.

        Args:
            path: path to the Zarr root directory
            simulator: simulator object
            sync_path: path for synchronization via file locks (files will be stored in the given path).
                It must differ from path, it must be accessible to all processes working on the store,
                and the underlying filesystem must support file locking.
        """
        if Path(path).exists():
            store = zarr.DirectoryStore(path)
            sync_path = sync_path or os.path.splitext(path)[0] + ".sync"
            return cls(
                zarr_store=store,
                simulator=simulator,
                sync_path=sync_path,
                from_scratch=False,
            )
        else:
            raise FileNotFoundError(f"There is no directory store at {path}.")

    def to_memory(self) -> "Store":
        """Make an in-memory copy of the existing Store using a Zarr MemoryStore."""
        memory_store = zarr.MemoryStore()
        zarr.convenience.copy_store(source=self._zarr_store, dest=memory_store)
        return Store(
            zarr_store=memory_store, simulator=self._simulator, from_scratch=False
        )



from importlib import import_module
from typing import Callable, Type, TypeVar, Union

import numpy as np
import torch
from toolz import compose
from toolz.dicttoolz import keyfilter
from torch.distributions import Normal, Uniform

from swyft.bounds import Bound, UnitCubeBound
from swyft.saveable import StateDictSaveable
from swyft.types import Array
from swyft.utils import array_to_tensor, tensor_to_array

PriorType = TypeVar("PriorType", bound="Prior")
PriorTruncatorType = TypeVar("PriorTruncatorType", bound="PriorTruncator")


class PriorTruncator(StateDictSaveable):
    """Samples from a truncated version of the prior and calculates the log_prob.

    Args:
        prior: Parameter prior
        bound: Bound object

    .. note::
        The prior truncator is defined through a swyft.Bound object, which
        sample from (subregions of) the hypercube, with swyft.Prior, which maps
        the samples onto parameters of interest.
    """

    def __init__(self, prior: "Prior", bound: Bound) -> None:
        """Instantiate prior truncator (combination of prior and bound).

        Args:
            prior: Prior object.
            bound: Bound on hypercube.  Set 'None' for untruncated priors.
        """
        self.prior = prior
        if bound is None:
            bound = UnitCubeBound(prior.n_parameters)
        self.bound = bound

    @property
    def cdf(self) -> Callable:
        return self.prior.cdf

    @property
    def icdf(self) -> Callable:
        return self.prior.icdf

    @property
    def n_parameters(self) -> int:
        return self.prior.n_parameters

    def sample(self, n_samples: int) -> np.ndarray:
        """Sample from truncated prior.

        Args:
            n_samples: Number of samples to return

        Returns:
            Samples: (n_samples, n_parameters)
        """
        u = self.bound.sample(n_samples)
        return self.prior.icdf(u)

    def log_prob(self, v: np.ndarray) -> np.ndarray:
        """Evaluate log probability.

        Args:
            v: (N, n_parameters) parameter points.

        Returns:
            log_prob: (N,)
        """
        u = self.prior.cdf(v)
        b = np.where(u.sum(axis=-1) == np.inf, 0.0, self.bound(u))
        log_prob = np.where(
            b == 0.0,
            -np.inf,
            self.prior.log_prob(v).sum(axis=-1) - np.log(self.bound.volume),
        )
        return log_prob

    def state_dict(self) -> dict:
        return dict(prior=self.prior.state_dict(), bound=self.bound.state_dict())

    @classmethod
    def from_state_dict(cls, state_dict: dict) -> PriorTruncatorType:
        prior = Prior.from_state_dict(state_dict["prior"])
        bound = Bound.from_state_dict(state_dict["bound"])
        return cls(prior, bound)


class InterpolatedTabulatedDistribution:
    def __init__(self, icdf: Callable, n_parameters: int, n_grid_points: int) -> None:
        r"""Create a distribution based off of a icdf. The distribution is defined by interpolating grid points.

        Args:
            icdf: inverse cumulative density function, aka ppf and uv
            n_parameters: number of parameters, dimensionality of the prior
            n_grid_points: number of grid points

        .. warning::
            Internally the mapping u -> v is tabulated on a linear grid on the
            interval [0, 1], with `n` grid points. In extreme cases, this can
            lead to approximation errors that can be mitigated by increasing
            `n`.
        """
        self.n_parameters = n_parameters
        self._grid = np.linspace(0, 1.0, n_grid_points)
        self._table = self._generate_table(icdf, self._grid, n_parameters)

    @staticmethod
    def _generate_table(
        uv: Callable, grid: np.ndarray, n_parameters: int
    ) -> np.ndarray:
        table = []
        for x in grid:
            table.append(uv(np.ones(n_parameters) * x))
        return np.array(table).T

    def cdf(self, v: np.ndarray) -> np.ndarray:
        """Map onto hypercube: v -> u

        Args:
            v: (N, n_parameters) physical parameter array

        Returns:
            u: (N, n_parameters) hypercube parameter array
        """
        u = np.empty_like(v)
        for i in range(self.n_parameters):
            u[:, i] = np.interp(
                v[:, i], self._table[i], self._grid, left=np.inf, right=np.inf
            )
        return u

    def icdf(self, u: np.ndarray) -> np.ndarray:
        """Map from hypercube: u -> v

        Args:
            u: (N, n_parameters) hypercube parameter array

        Returns:
            v: (N, n_parameters) physical parameter array
        """
        v = np.empty_like(u)
        for i in range(self.n_parameters):
            v[:, i] = np.interp(
                u[:, i], self._grid, self._table[i], left=np.inf, right=np.inf
            )
        return v

    def log_prob(self, v: np.ndarray, du: float = 1e-6) -> np.ndarray:
        """Log probability.

        Args:
            v: (N, n_parameters) physical parameter array
            du: Step-size of numerical derivatives

        Returns:
            log_prob: (N, n_parameters) factors of pdf
        """
        dv = np.empty_like(v)
        u = self.cdf(v)
        for i in range(self.n_parameters):
            dv[:, i] = np.interp(
                u[:, i] + (du / 2), self._grid, self._table[i], left=None, right=None
            )
            dv[:, i] -= np.interp(
                u[:, i] - (du / 2), self._grid, self._table[i], left=None, right=None
            )
        log_prob = np.where(u == np.inf, -np.inf, np.log(du) - np.log(dv + 1e-300))
        return log_prob


# TODO this could be improved with some thought
# it merely wraps a torch distribution and keeps track of the arguments...
class Prior(StateDictSaveable):
    def __init__(
        self, cdf: Callable, icdf: Callable, log_prob: Callable, n_parameters: int
    ) -> None:
        r"""Fully factorizable prior.

        Args:
            cdf: cumulative density function, aka vu
            icdf: inverse cumulative density function, aka ppf and uv
            log_prob: log density function
            n_parameters: number of parameters / dimensionality of the prior

        .. note::
            The prior is defined through the mapping :math:`u\to v`, from the
            Uniform distribution, :math:`u\sim \text{Unif}(0, 1)` onto the
            parameters of interest, :math:`v`.  This mapping corresponds to the
            inverse cummulative distribution function, and is internally used
            to perform inverse transform sampling.  Sampling happens in the
            swyft.Bound object.
        """
        self.cdf = cdf
        self.icdf = icdf
        self.log_prob = log_prob
        self.n_parameters = n_parameters
        self.method = "__init__"
        self._state_dict = {
            "method": self.method,
            "cdf": self.cdf,
            "icdf": self.icdf,
            "log_prob": self.log_prob,
            "n_parameters": self.n_parameters,
        }
        self.distribution = None

    @classmethod
    def from_torch_distribution(
        cls: Type[PriorType],
        distribution: torch.distributions.Distribution,
    ) -> PriorType:
        r"""Create a prior from a batched pytorch distribution.

        For example, ``distribution = torch.distributions.Uniform(-1 * torch.ones(5), 1 * torch.ones(5))``.

        Args:
            distribution: pytorch distribution

        Returns:
            Prior
        """
        assert (
            len(distribution.batch_shape) == 1
        ), f"{distribution.batch_shape} must be one dimensional"
        assert (
            len(distribution.event_shape) == 0
        ), f"{distribution} must be factorizable and report the log_prob of every dimension (i.e. all dims are in batch_shape)"
        prior = cls(
            cdf=compose(tensor_to_array, distribution.cdf, array_to_tensor),
            icdf=compose(tensor_to_array, distribution.icdf, array_to_tensor),
            log_prob=compose(tensor_to_array, distribution.log_prob, array_to_tensor),
            n_parameters=distribution.batch_shape.numel(),
        )
        prior.distribution = distribution
        prior.method = "from_torch_distribution"
        prior._state_dict = {
            "method": prior.method,
            "name": distribution.__class__.__name__,
            "module": distribution.__module__,
            "kwargs": keyfilter(
                lambda x: x in distribution.__class__.arg_constraints,
                distribution.__dict__,  # this depends on all relevant arguments being contained with prior.distribution.__class__.arg_constraints
            ),
        }
        return prior

    @classmethod
    def from_uv(
        cls, icdf: Callable, n_parameters: int, n_grid_points: int = 10_000
    ) -> PriorType:
        """Create a prior which depends on ``InterpolatedTabulatedDistribution``, i.e. an interpolated representation of the icdf, cdf, and log_prob.

        .. warning::
            Internally the mapping u -> v is tabulated on a linear grid on the
            interval [0, 1], with `n` grid points. In extreme cases, this can
            lead to approximation errors that can be mitigated by increasing
            `n` (in some cases).

        Args:
            icdf: map from hypercube: u -> v. inverse cumulative density function (icdf)
            n_parameters: number of parameters / dimensionality of the prior
            n_grid_points: number of grid points from which to interpolate the icdf, cdf, and log_prob

        Returns:
            Prior
        """
        raise NotImplementedError("This was too inaccurate.")
        distribution = InterpolatedTabulatedDistribution(
            icdf, n_parameters, n_grid_points
        )
        prior = cls(
            cdf=distribution.v,
            icdf=distribution.u,
            log_prob=distribution.log_prob,
            n_parameters=n_parameters,
        )
        prior.distribution = distribution
        prior.method = "from_from_uv"
        prior.state_dict = None  # TODO, make like above.
        return prior

    def state_dict(self) -> dict:
        return self._state_dict

    @classmethod
    def from_state_dict(cls, state_dict: dict) -> PriorType:
        method = state_dict["method"]

        if method == "__init__":
            kwargs = keyfilter(lambda x: x != "method", state_dict)
            return cls(**kwargs)
        elif method == "from_torch_distribution":
            name = state_dict["name"]
            module = state_dict["module"]
            kwargs = state_dict["kwargs"]
            distribution = getattr(import_module(module), name)
            distribution = distribution(**kwargs)
            return getattr(cls, method)(distribution)
        else:
            NotImplementedError()


def get_uniform_prior(low: Array, high: Array) -> Prior:
    distribution = Uniform(array_to_tensor(low), array_to_tensor(high))
    return Prior.from_torch_distribution(distribution)


def get_diagonal_normal_prior(loc: Array, scale: Array) -> Prior:
    distribution = Normal(array_to_tensor(loc), array_to_tensor(scale))
    return Prior.from_torch_distribution(distribution)


if __name__ == "__main__":
    pass



import matplotlib.patches as mpatches
import matplotlib.path as mpath

from swyft.plot.histogram import split_corner_axes
from swyft.utils.marginals import get_d_dim_marginal_indices


def diagonal_constraint(axes, bounds, alpha=0.25):
    _, diag, _ = split_corner_axes(axes)
    for i, ax in enumerate(diag):
        xlim = ax.get_xlim()
        ax.axvspan(xlim[0], bounds[i, 0], alpha=alpha)
        ax.axvspan(bounds[i, 1], xlim[1], alpha=alpha)
    return axes


def mask_outside_polygon(poly_verts, ax, facecolor=None, edgecolor=None, alpha=0.25):
    """
    Plots a mask on the specified axis ("ax", defaults to plt.gca()) such that
    all areas outside of the polygon specified by "poly_verts" are masked.

    "poly_verts" must be a list of tuples of the verticies in the polygon in
    counter-clockwise order.

    Returns the matplotlib.patches.PathPatch instance plotted on the figure.
    """

    # Get current plot limits
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    # Verticies of the plot boundaries in clockwise order
    bound_verts = [
        (xlim[0], ylim[0]),
        (xlim[0], ylim[1]),
        (xlim[1], ylim[1]),
        (xlim[1], ylim[0]),
        (xlim[0], ylim[0]),
    ]

    # A series of codes (1 and 2) to tell matplotlib whether to draw a line or
    # move the "pen" (So that there's no connecting line)
    bound_codes = [mpath.Path.MOVETO] + (len(bound_verts) - 1) * [mpath.Path.LINETO]
    poly_codes = [mpath.Path.MOVETO] + (len(poly_verts) - 1) * [mpath.Path.LINETO]

    # Plot the masking patch
    path = mpath.Path(bound_verts + poly_verts, bound_codes + poly_codes)
    patch = mpatches.PathPatch(
        path, facecolor=facecolor, edgecolor=edgecolor, alpha=alpha
    )
    patch = ax.add_patch(patch)

    # Reset the plot limits to their original extents
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    return patch


def construct_vertices(bounds, x, y):
    xbound = bounds[x, :]
    ybound = bounds[y, :]
    return [
        (xbound[0], ybound[0]),
        (xbound[0], ybound[1]),
        (xbound[1], ybound[1]),
        (xbound[1], ybound[0]),
        (xbound[0], ybound[0]),
    ][
        ::-1
    ]  # counter clockwise


def lower_constraint(axes, bounds, alpha=0.25):
    d, _ = axes.shape
    upper_inds = get_d_dim_marginal_indices(d, 2)
    for i in upper_inds:
        x, y = i
        ax = axes[y, x]  # targets the lower left corner
        inside_verts = construct_vertices(bounds, x, y)
        mask_outside_polygon(inside_verts, ax, alpha=alpha)



from swyft.utils.array import (
    all_finite,
    array_to_tensor,
    dict_to_device,
    tensor_to_array,
)
from swyft.utils.marginals import (
    get_corner_marginal_indices,
    get_d_dim_marginal_indices,
    tupleize_marginal_indices,
)
from swyft.utils.misc import depth, is_cuda_available, is_empty

__all__ = [
    "array_to_tensor",
    "all_finite",
    "depth",
    "dict_to_device",
    "is_cuda_available",
    "is_empty",
    "get_d_dim_marginal_indices",
    "get_corner_marginal_indices",
    "tensor_to_array",
    "tupleize_marginal_indices",
]



from typing import (
    Callable,
    Dict,
    Hashable,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np
import torch
import torch.nn as nn
from toolz.dicttoolz import valmap
from torch.utils.data import DataLoader, Dataset, random_split

import swyft
import swyft.utils
from swyft.saveable import StateDictSaveable, StateDictSaveableType
from swyft.types import Array, Device, MarginalIndex, MarginalToArray, ObsType, PathType
from swyft.utils.array import array_to_tensor, dict_array_to_tensor
from swyft.utils.marginals import tupleize_marginal_indices

SchedulerType = Union[
    torch.optim.lr_scheduler._LRScheduler, torch.optim.lr_scheduler.ReduceLROnPlateau
]


def split_length_by_percentage(length: int, percents: Sequence[float]) -> Sequence[int]:
    """Given the length of a sequence, return the indices which would divide it into `percents` parts.
    Any rounding errors go into the first part.

    Args:
        length
        percents

    Returns:
        length_of_parts
    """
    assert np.isclose(sum(percents), 1.0), f"{percents} does not sum to 1."
    lengths = [int(percent * length) for percent in percents]

    # Any extra from round off goes to the first split.
    difference = length - sum(lengths)
    lengths[0] += difference
    assert length == sum(
        lengths
    ), f"Splitting into {lengths} should equal total {length}."
    return lengths


def get_ntrain_nvalid(
    validation_amount: Union[float, int], len_dataset: int
) -> Tuple[int, int]:
    """Divide a dataset into a training and validation set.

    Args:
        validation_amount: percentage or number of elements in the validation set
        len_dataset: total length of the dataset

    Raises:
        TypeError: When the validation_amount is neither a float or int.

    Returns:
        (n_train, n_valid)
    """
    assert validation_amount > 0
    if isinstance(validation_amount, float):
        percent_validation = validation_amount
        percent_train = 1.0 - percent_validation
        n_valid, n_train = split_length_by_percentage(
            len_dataset, (percent_validation, percent_train)
        )
        if n_valid % 2 != 0:
            n_valid += 1
            n_train -= 1
    elif isinstance(validation_amount, int):
        n_valid = validation_amount
        n_train = len_dataset - n_valid
        assert n_train > 0

        if n_valid % 2 != 0:
            n_valid += 1
            n_train -= 1
    else:
        raise TypeError("validation_amount must be int or float")
    return n_train, n_valid


def double_observation(f: torch.Tensor) -> torch.Tensor:
    """Double observation vector as (A, B, C, D) --> (A, A, B, B, C, C, D, D)

    Args:
        f: Observation vectors (n_batch, ...)

    Returns:
        Observation vectors (2*n_batch, ...)
    """
    return torch.repeat_interleave(f, 2, dim=0)


def double_parameters(parameters: torch.Tensor) -> torch.Tensor:
    """Double parameters as (A, B, C, D) --> (A, B, A, B, C, D, C, D) etc

    Args:
        parameters: Parameter vectors (n_batch, n_parameters).

    Returns:
        parameters with shape (2*n_batch, n_parameters).
    """
    n_batch, n_parameters = parameters.shape
    assert n_batch % 2 == 0, "n_batch must be divisible by two."
    out = torch.repeat_interleave(parameters.view(-1, 2 * n_parameters), 2, dim=0).view(
        -1, n_parameters
    )
    return out


MarginalRatioEstimatorType = TypeVar(
    "MarginalRatioEstimatorType", bound="MarginalRatioEstimator"
)


class MarginalRatioEstimator(StateDictSaveable):
    """Handles the training and evaluation of a ratio estimator. Which ratios are defined by the `marginal_indices` attribute.
    The network must take observation dictionaries and parameter arrays and produce estimated an `log_ratio` for every marginal of interest.
    """

    def __init__(
        self,
        marginal_indices: MarginalIndex,
        network: nn.Module,
        device: Device,
    ) -> None:
        """Define the marginals of interest with `marginal_indices` and the estimator architechture with `network`.

        Args:
            marginal_indices: marginals of interest defined by the parameter index
            network: a neural network which accepts `observation` and `parameters` and returns `len(marginal_indices)` ratios.
            device
        """
        self.marginal_indices = tupleize_marginal_indices(marginal_indices)
        self.device = device
        self.network = network
        self.network.to(self.device)

        self.epoch = None
        self.best_network_state_dict = None
        self.min_loss = float("-Inf")
        self.optimizer = None
        self.scheduler = None

    def train(
        self,
        dataset: Dataset,
        batch_size: int = 50,
        learning_rate: float = 5e-4,
        validation_percentage: float = 0.1,
        optimizer: Callable = torch.optim.Adam,
        scheduler: Optional[Callable] = torch.optim.lr_scheduler.ReduceLROnPlateau,
        scheduler_kwargs: dict = {"factor": 0.1, "patience": 5},
        early_stopping_patience: Optional[int] = 25,
        max_epochs: int = 2 ** 31 - 1,
        nworkers: int = 0,
        non_blocking: bool = True,
        pin_memory: bool = True,
    ) -> None:
        """Train the ratio estimator based off of a `dataset` containing observation and parameter pairs.

        Note: if the network has already been trained, training will resume where it left off.
        This effectively ignores `optimizer`, `learning_rate`, `scheduler`, and `scheduler_args`.

        Args:
            dataset: torch dataset which returns a tuple of (`observation`, `parameters`)
            batch_size
            learning_rate
            validation_percentage: Approximates the percentage of `dataset` used in the validation set
            optimizer: from `torch.optim` optimizer. It can only accept two arguments: `parameters` and `lr`. Need more arguments? Use `functools.partial`.
            scheduler: from `torch.optim.lr_scheduler`
            scheduler_kwargs: The arguments which get passed to `scheduler`
            early_stopping_patience: after this many fuitless epochs, training stops
            max_epochs: maximum number of epochs to train
            nworkers: number of workers to divide `dataloader` duties between. 0 implies one thread for training and dataloading.
            non_blocking: consult torch documentation, generally use `True`
            pin_memory: consult torch documentation, generally use `True`
        """
        if early_stopping_patience is None:
            early_stopping_patience = max_epochs

        if self.optimizer is None:
            self.optimizer = optimizer(self.network.parameters(), lr=learning_rate)

        if scheduler is not None and self.scheduler is None:
            self.scheduler = scheduler(self.optimizer, **scheduler_kwargs)

        n_train, n_valid = get_ntrain_nvalid(validation_percentage, len(dataset))
        dataset_train, dataset_valid = random_split(dataset, [n_train, n_valid])
        train_loader = DataLoader(
            dataset_train,
            batch_size=batch_size,
            num_workers=nworkers,
            pin_memory=pin_memory,
            drop_last=True,
        )
        valid_loader = DataLoader(
            dataset_valid,
            batch_size=min(batch_size, n_valid),
            num_workers=nworkers,
            pin_memory=pin_memory,
            drop_last=True,
        )

        n_validation_batches = len(valid_loader) if len(valid_loader) != 0 else 1
        validation_losses = []
        self.epoch, fruitless_epoch = 0, 0
        while self.epoch < max_epochs and fruitless_epoch < early_stopping_patience:
            # Training
            self.network.train()
            for observation, _, v in train_loader:
                self.optimizer.zero_grad()
                observation = swyft.utils.dict_to_device(
                    observation, device=self.device, non_blocking=non_blocking
                )
                v = v.to(self.device)
                loss = self._loss(observation, v).sum(dim=0)
                loss.backward()
                self.optimizer.step()

            self.epoch += 1

            # Evaluation
            self.network.eval()
            loss_sum = 0
            with torch.no_grad():
                for observation, _, v in valid_loader:
                    observation = swyft.utils.dict_to_device(
                        observation, device=self.device, non_blocking=non_blocking
                    )
                    v = v.to(self.device)
                    validation_loss = self._loss(observation, v).sum(dim=0)
                    loss_sum += validation_loss
                loss_avg = loss_sum / n_validation_batches
                print(
                    "\rtraining: lr=%.2g, epoch=%i, validation loss=%.4g"
                    % (self._get_last_lr(self.scheduler), self.epoch, loss_avg),
                    end="",
                    flush=True,
                )
                validation_losses.append(loss_avg)

                if self.epoch == 0 or self.min_loss > validation_loss:
                    fruitless_epoch = 0
                    self.min_loss = validation_loss
                    self.best_network_state_dict = self.network.state_dict()
                else:
                    fruitless_epoch += 1

                if self.scheduler is not None:
                    self.scheduler.step(loss_avg)
        print("")

    def log_ratio(
        self,
        observation: ObsType,
        v: Array,
        batch_size: Optional[int] = None,
    ) -> MarginalToArray:
        """Evaluate the ratio estimator on a single `observation` with many `parameters`.
        The `parameters` correspond to `v`, i.e. the "physical" parameterization.
        (As opposed to `u` which is mapped to the hypercube.)

        Args:
            observation: a single observation to estimate ratios on (Cannot have a batch dimension!)
            v: parameters
            batch_size: divides the evaluation into batches of this size

        Returns:
            MarginalToArray: the ratios of each marginal in `marginal_indices`. Each marginal index is a key.
        """
        was_training = self.network.training
        self.network.eval()

        with torch.no_grad():
            observation = dict_array_to_tensor(observation, device=self.device)
            features = self.network.head(
                {key: value.unsqueeze(0) for key, value in observation.items()}
            )
            len_v = len(v)
            if batch_size is None or len_v <= batch_size:
                v = array_to_tensor(v, device=self.device)
                repeated_features = features.expand(v.size(0), *features.shape[1:])
                ratio = self.network.tail(repeated_features, v).cpu().numpy()
            else:
                ratio = []
                for i in range(len_v // batch_size + 1):
                    parameter_batch = array_to_tensor(
                        v[i * batch_size : (i + 1) * batch_size, :],
                        device=self.device,
                    )
                    feature_batch = features.expand(
                        parameter_batch.size(0), *features.shape[1:]
                    )
                    ratio_batch = (
                        self.network.tail(feature_batch, parameter_batch).cpu().numpy()
                    )
                    ratio.append(ratio_batch)
                ratio = np.vstack(ratio)

        if was_training:
            self.network.train()
        else:
            self.network.eval()

        return {k: ratio[..., i] for i, k in enumerate(self.marginal_indices)}

    @staticmethod
    def _repeat_observation_to_match_v(
        observation: Dict[Hashable, torch.Tensor], v: torch.Tensor
    ) -> Dict[Hashable, torch.Tensor]:
        b, *_ = v.size()
        return valmap(lambda x: x.unsqueeze(0).expand(b, *x.size()), observation)

    @staticmethod
    def _get_last_lr(scheduler: SchedulerType) -> float:
        """Get the last learning rate from a `lr_scheduler`."""
        if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
            if scheduler.best == float("Inf"):
                return scheduler.optimizer.param_groups[0]["lr"]
            else:
                return scheduler._last_lr[-1]
        elif isinstance(scheduler, torch.optim.lr_scheduler._LRScheduler):
            return scheduler.get_last_lr()
        else:
            raise NotImplementedError(
                f"Cannot determine learning_rate from {scheduler}"
            )

    def _loss(
        self, observation: Dict[Hashable, torch.Tensor], parameters: torch.Tensor
    ) -> torch.Tensor:
        """Calculate the marginal-wise losses.

        Args:
            observation: a batch of observations within a dictionary
            parameters: a batch of parameters

        Returns:
            torch.Tensor: the marginal-wise losses with `len(self.marginal_indices)`
        """
        n_batch = parameters.size(0)
        assert (
            n_batch % 2 == 0
        ), "Loss function can only handle even-numbered batch sizes."
        assert all(
            [value.size(0) == n_batch for value in observation.values()]
        ), "The observation batch_size must agree with the parameter batch_size."

        # Repeat interleave
        observation_doubled = valmap(double_observation, observation)
        parameters_doubled = double_parameters(parameters)

        lnL = self.network(observation_doubled, parameters_doubled)
        lnL = lnL.view(-1, 4, lnL.shape[-1])

        loss = -torch.nn.functional.logsigmoid(lnL[:, 0])
        loss += -torch.nn.functional.logsigmoid(-lnL[:, 1])
        loss += -torch.nn.functional.logsigmoid(-lnL[:, 2])
        loss += -torch.nn.functional.logsigmoid(lnL[:, 3])
        loss = loss.sum(axis=0) / n_batch

        return loss

    def state_dict(self) -> dict:
        return {
            "marginal_indices": self.marginal_indices,
            "network": self.network.state_dict(),
            "epoch": self.epoch,
            "min_loss": self.min_loss,
            "best_network_state_dict": self.best_network_state_dict,
            "optimizer": self.optimizer.state_dict()
            if self.optimizer is not None
            else None,
            "scheduler": self.scheduler.state_dict()
            if self.scheduler is not None
            else None,
        }  # TODO this could also save the __class__ and use importlib, thereby reducing the arguments to from_state_dict

    @classmethod
    def from_state_dict(
        cls,
        network: torch.nn.Module,
        optimizer: Optional[torch.optim.Optimizer],
        scheduler: Optional[SchedulerType],
        device: Device,
        state_dict: dict,
    ) -> MarginalRatioEstimatorType:
        """Instantiate a MarginalRatioEstimator from a state_dict, along with a few necessary python objects.

        Args:
            network: initialized network
            optimizer: same optimizer as used by saved model
            scheduler: same scheduler as used by saved model
            device
            state_dict

        Returns:
            MarginalRatioEstimatorType: loaded model
        """
        marginal_ratio_estimator = cls.__new__(cls)

        marginal_ratio_estimator.marginal_indices = state_dict["marginal_indices"]
        marginal_ratio_estimator.epoch = state_dict["epoch"]
        marginal_ratio_estimator.min_loss = state_dict["min_loss"]
        marginal_ratio_estimator.best_network_state_dict = state_dict[
            "best_network_state_dict"
        ]
        marginal_ratio_estimator.device = device

        marginal_ratio_estimator.network = network
        marginal_ratio_estimator.optimizer = optimizer
        marginal_ratio_estimator.scheduler = scheduler

        marginal_ratio_estimator.network.load_state_dict(state_dict["network"])

        if optimizer is not None and state_dict["optimizer"] is None:
            raise FileNotFoundError(
                "There was no data about the optimizer in the state_dict"
            )
        elif optimizer is not None and state_dict["optimizer"] is not None:
            marginal_ratio_estimator.optimizer.load_state_dict(state_dict["optimizer"])
        else:
            pass

        if scheduler is not None and state_dict["scheduler"] is None:
            raise FileNotFoundError(
                "There was no data about the scheduler in the state_dict"
            )
        elif scheduler is not None and state_dict["scheduler"] is not None:
            marginal_ratio_estimator.scheduler.load_state_dict(state_dict["scheduler"])
        else:
            pass

        return marginal_ratio_estimator

    @classmethod
    def load(
        cls: Type[StateDictSaveableType],
        network: torch.nn.Module,
        device: Device,
        filename: PathType,
        optimizer: Optional[torch.optim.Optimizer] = None,
        scheduler: Optional[SchedulerType] = None,
    ) -> StateDictSaveableType:
        sd = torch.load(filename)
        return cls.from_state_dict(
            network=network,
            optimizer=optimizer,
            scheduler=scheduler,
            device=device,
            state_dict=sd,
        )


if __name__ == "__main__":
    pass



from swyft.inference.marginalposterior import MarginalPosterior
from swyft.inference.marginalratioestimator import MarginalRatioEstimator

__all__ = [
    "MarginalRatioEstimator",
    "MarginalPosterior",
]



import tempfile
from itertools import product
from pathlib import Path

import numpy as np
import pytest

import swyft
from swyft.bounds import RectangleBound, UnitCubeBound
from swyft.prior import PriorTruncator, get_diagonal_normal_prior, get_uniform_prior


class TestSaveLoadPriorTruncator:
    dimension = 10
    n_samples = 1_000
    rectangle_bounds = np.stack([np.zeros(dimension), np.ones(dimension)], axis=-1)

    distributions = [
        get_uniform_prior(low=-1 * np.ones(dimension), high=2 * np.ones(dimension)),
        get_diagonal_normal_prior(
            loc=np.random.randn(dimension), scale=np.ones(dimension)
        ),
    ]
    bounds = [
        None,
        UnitCubeBound(dimension),
        RectangleBound(rectangle_bounds),
    ]

    @classmethod
    def setup_class(cls):
        cls.directory = tempfile.TemporaryDirectory()

    @classmethod
    def teardown_class(cls):
        cls.directory.cleanup()

    @pytest.mark.parametrize("prior, bound", product(distributions, bounds))
    def test_save_load_prior_truncator_no_bound(
        self, prior: swyft.Prior, bound: swyft.Bound
    ):
        prior_truncator = PriorTruncator(prior, bound=None)

        # Saving
        path = Path(self.directory.name) / f"prior_truncator_{bound}"
        prior_truncator.save(path)

        # Loading
        prior_truncator_loaded = PriorTruncator.load(path)

        # Testing by cdf
        # (icdf or log_prob would also be fine.)
        samples = prior_truncator.sample(self.n_samples)

        log_prob_true = prior_truncator.log_prob(samples)
        log_prob_esti = prior_truncator_loaded.log_prob(samples)
        assert np.allclose(log_prob_true, log_prob_esti)


if __name__ == "__main__":
    pass



from typing import Dict, Hashable, Optional, Union

import numpy as np
import torch

from swyft.types import Array, Device


def dict_to_device(d, device, non_blocking=False):
    return {k: v.to(device, non_blocking=non_blocking) for k, v in d.items()}


def dict_array_to_tensor(
    d, device="cpu", non_blocking=False, indices=slice(0, None)
) -> Dict[Hashable, torch.Tensor]:
    return {
        k: array_to_tensor(v[indices]).to(device, non_blocking=non_blocking)
        for k, v in d.items()
    }


np_bool_types = [bool]
np_int_types = [np.int8, np.int16, np.int32, np.int64]
np_float_types = [np.float32, np.float64]
np_complex_types = [np.complex64, np.complex128]
torch_bool_types = [torch.bool]
torch_int_types = [torch.int8, torch.int16, torch.int32, torch.int64]
torch_float_types = [torch.float32, torch.float64]
torch_complex_types = [torch.complex64, torch.complex128]


def array_to_tensor(
    array: Array, dtype: Optional[torch.dtype] = None, device: Optional[Device] = None
) -> torch.Tensor:
    """Converts np.ndarray and torch.Tensor to torch.Tensor with dtype and on device.
    When dtype is None, unsafe casts all float-type arrays to torch.float32 and all int-type arrays to torch.int64
    """
    if not isinstance(array, (np.ndarray, torch.Tensor)):
        array = np.asarray(array)

    input_dtype = array.dtype
    if isinstance(input_dtype, np.dtype):
        if dtype is None:
            if input_dtype in np_float_types:
                dtype = torch.float32
            elif input_dtype in np_int_types:
                dtype = torch.int64
            elif input_dtype in np_bool_types:
                dtype = torch.bool
            elif input_dtype in np_complex_types:
                dtype = torch.complex64
            else:
                raise TypeError(
                    f"{input_dtype} was not a supported numpy int, float, bool, or complex."
                )
        return torch.from_numpy(array).to(dtype=dtype, device=device)
    elif isinstance(input_dtype, torch.dtype):
        if dtype is None:
            if input_dtype in torch_float_types:
                dtype = torch.float32
            elif input_dtype in torch_int_types:
                dtype = torch.int64
            elif input_dtype in torch_bool_types:
                dtype = torch.bool
            elif input_dtype in torch_complex_types:
                dtype = torch.complex64
            else:
                raise TypeError(
                    f"{input_dtype} was not a supported torch int, float, bool, or complex."
                )
        return array.to(dtype=dtype, device=device)
    else:
        raise TypeError(
            f"{input_dtype} was not recognized as a supported numpy.dtype or torch.dtype."
        )


def tensor_to_array(
    tensor: Array, dtype: Optional[np.dtype] = None, copy: bool = True
) -> np.ndarray:
    if isinstance(tensor, torch.Tensor):
        out = np.asarray(tensor.detach().cpu().numpy(), dtype=dtype)
    else:
        out = np.asarray(tensor, dtype=dtype)

    if copy:
        return out.copy()
    else:
        return out


def tobytes(x: Array) -> Array:
    if isinstance(x, np.ndarray):
        return x.tobytes()
    elif isinstance(x, torch.Tensor):
        return x.numpy().tobytes()
    else:
        raise TypeError(f"{type(x)} does not support tobytes.")


def _all_finite(x: Array) -> Array:
    if isinstance(x, torch.Tensor):
        return torch.all(torch.isfinite(x))
    else:
        return np.all(np.isfinite(x))


def all_finite(x: Union[dict, torch.Tensor, np.ndarray, list]) -> bool:
    if isinstance(x, dict):
        return all(_all_finite(v) for v in x.values())
    elif isinstance(x, (torch.Tensor, np.ndarray)):
        return _all_finite(x)
    elif isinstance(x, list):
        return all(_all_finite(v) for v in x)
    else:
        raise NotImplementedError("That type is not yet implemented.")



from swyft.plot.constraint import diagonal_constraint, lower_constraint
from swyft.plot.histogram import corner, hist1d
from swyft.plot.mass import empirical_z_score_corner, plot_empirical_z_score
from swyft.plot.violin import violin

__all__ = [
    "corner",
    "diagonal_constraint",
    "hist1d",
    "lower_constraint",
    "plot_empirical_z_score",
    "empirical_z_score_corner",
    "violin",
]



import tempfile
from itertools import product
from pathlib import Path
from typing import Dict, Hashable, Optional

import numpy as np
import pytest
import torch

import swyft.inference.marginalratioestimator as mre
import swyft.networks.classifier as classifier
from swyft.bounds import Bound, RectangleBound
from swyft.inference.marginalposterior import MarginalPosterior
from swyft.prior import (
    Prior,
    PriorTruncator,
    get_diagonal_normal_prior,
    get_uniform_prior,
)
from swyft.store.dataset import SimpleDataset
from swyft.types import MarginalIndex
from swyft.utils import tupleize_marginal_indices


class AllOneNetwork(torch.nn.Module, classifier.HeadTailClassifier):
    def __init__(self, marginal_indices: MarginalIndex) -> None:
        super().__init__()
        self.marginal_indices = tupleize_marginal_indices(marginal_indices)
        self.block_shape = classifier.ParameterTransform.get_marginal_block_shape(
            self.marginal_indices
        )
        self.m, _ = self.block_shape
        self.register_buffer("out", torch.ones(1, self.m))

    def forward(self, _, y: torch.Tensor) -> torch.Tensor:
        b, *_ = y.shape
        return self.out.expand(b, self.m)

    def head(self, _: Dict[Hashable, torch.Tensor]) -> torch.Tensor:
        return self.out

    def tail(self, features: torch.Tensor, parameters: torch.Tensor) -> torch.Tensor:
        return self(features, parameters)


class TestMarginalPosterior:
    @classmethod
    def setup_class(cls):
        cls.observation_key = "x"
        cls.observation_shapes = {cls.observation_key: (10,)}
        cls.n_parameters = 2
        cls.device = "cpu"

    @classmethod
    def get_marginal_ratio_estimator(cls, marginal_indices: MarginalIndex):
        network = classifier.get_marginal_classifier(
            observation_key=cls.observation_key,
            marginal_indices=marginal_indices,
            observation_shapes=cls.observation_shapes,
            n_parameters=cls.n_parameters,
            hidden_features=16,
            num_blocks=1,
        )
        marginal_ratio_estimator = mre.MarginalRatioEstimator(
            marginal_indices=marginal_indices,
            network=network,
            device=cls.device,
        )
        return marginal_ratio_estimator

    def test_init_prior(self):
        n_parameters = 2
        prior = get_uniform_prior([-0.5] * n_parameters, [5.0] * n_parameters)

        marginal_indices = list(range(n_parameters))
        mre = self.get_marginal_ratio_estimator(marginal_indices)

        mp = MarginalPosterior(mre, prior)
        assert isinstance(mp.prior, Prior)

    def test_init_prior_truncator(self):
        bound = [[0.1, 0.9], [0.2, 0.8]]
        bound = np.asarray(bound)
        rectangle_bound = RectangleBound(bound)

        n_parameters = len(bound)
        prior = get_uniform_prior([-0.5] * n_parameters, [5.0] * n_parameters)

        marginal_indices = list(range(n_parameters))
        mre = self.get_marginal_ratio_estimator(marginal_indices)

        mp = MarginalPosterior(mre, prior, rectangle_bound)
        assert isinstance(mp.prior_truncator, PriorTruncator)

    @pytest.mark.parametrize("marginal_indices", [[0], [0, 1], [(0, 1)]])
    def test_truncate(self, marginal_indices: MarginalIndex):
        prior = get_uniform_prior([-2.5] * self.n_parameters, [5.0] * self.n_parameters)
        marginal_indices = list(range(self.n_parameters))
        network = AllOneNetwork(marginal_indices)
        marginal_ratio_estimator = mre.MarginalRatioEstimator(
            marginal_indices=marginal_indices,
            network=network,
            device=self.device,
        )
        mp = MarginalPosterior(marginal_ratio_estimator, prior)

        n_samples = 1_000
        fabricated_observation = {
            key: torch.rand(*shape) for key, shape in self.observation_shapes.items()
        }
        bound = mp.truncate(
            n_samples,
            fabricated_observation,
            threshold=-50.0,
        )  # TODO make this test the actual behavior of the function.
        assert isinstance(bound, Bound)

    @pytest.mark.parametrize("marginal_indices", [[0], [0, 1], [(0, 1)]])
    def test_empirical_mass(self, marginal_indices):
        prior = get_uniform_prior([-2.5] * self.n_parameters, [5.0] * self.n_parameters)
        marginal_indices = list(range(self.n_parameters))
        network = AllOneNetwork(marginal_indices)
        marginal_ratio_estimator = mre.MarginalRatioEstimator(
            marginal_indices=marginal_indices,
            network=network,
            device=self.device,
        )
        mp = MarginalPosterior(marginal_ratio_estimator, prior)

        n_observations = 100
        us = torch.randn(n_observations, self.n_parameters)
        vs = torch.randn(n_observations, self.n_parameters)
        observations = {
            key: torch.rand(n_observations, *shape)
            for key, shape in self.observation_shapes.items()
        }
        dataset = SimpleDataset(observations, us, vs)

        empirical_mass, _ = mp.empirical_mass(
            n_observations=n_observations,
            n_posterior_samples=1000,
            dataset=dataset,
        )  # TODO make this test the actual behavior of the function.
        assert isinstance(empirical_mass, dict)

    @pytest.mark.skip
    def test_log_prob_value_with_fake_mre(self):
        raise NotImplementedError(
            "This test could be constructed using a conjugate distribution and a 'decoy' marginal ratio estimator."
        )

    @pytest.mark.parametrize(
        "marginal_indices, batch_size",
        product(
            [[0], [0, 1], [(0, 1)]],
            [None, 10],
        ),
    )
    def test_log_prob_shape(
        self, marginal_indices: MarginalIndex, batch_size: Optional[int]
    ):
        n_batch = 100
        marginal_indices = tupleize_marginal_indices(marginal_indices)
        marginal_ratio_estimator = self.get_marginal_ratio_estimator(marginal_indices)

        prior = get_diagonal_normal_prior(
            loc=[0] * self.n_parameters, scale=[1] * self.n_parameters
        )

        marginal_posterior = MarginalPosterior(marginal_ratio_estimator, prior)

        fabricated_observation = {
            key: torch.rand(*shape) for key, shape in self.observation_shapes.items()
        }
        fabricated_v = torch.randn(n_batch, self.n_parameters)
        log_prob = marginal_posterior.log_prob(
            observation=fabricated_observation,
            v=fabricated_v,
            batch_size=batch_size,
        )
        assert set(log_prob.keys()) == set(marginal_indices)
        for _, value in log_prob.items():
            assert value.shape == (n_batch,)

    @pytest.mark.parametrize(
        "n_samples, marginal_indices, batch_size",
        product(
            [101, 1000],
            [[0], [0, 1], [(0, 1)]],
            [None, 10],
        ),
    )
    def test_weighted_sample_shape(
        self, n_samples: int, marginal_indices: MarginalIndex, batch_size: Optional[int]
    ):
        marginal_indices = tupleize_marginal_indices(marginal_indices)
        n_marginal_parameters = len(marginal_indices[0])
        marginal_ratio_estimator = self.get_marginal_ratio_estimator(marginal_indices)

        prior = get_diagonal_normal_prior(
            loc=[0] * self.n_parameters, scale=[1] * self.n_parameters
        )

        marginal_posterior = MarginalPosterior(marginal_ratio_estimator, prior)

        fabricated_observation = {
            key: torch.rand(*shape) for key, shape in self.observation_shapes.items()
        }
        weighted_samples = marginal_posterior.weighted_sample(
            n_samples=n_samples,
            observation=fabricated_observation,
            batch_size=batch_size,
        )

        assert set(weighted_samples.marginal_indices) == set(marginal_indices)
        assert weighted_samples.v.shape == (
            n_samples,
            self.n_parameters,
        ), "size of the parameters v does not match"
        for marginal_index in marginal_indices:
            logweight, marginal = weighted_samples.get_logweight_marginal(
                marginal_index
            )
            assert logweight.shape == (n_samples,)
            assert marginal.shape == (
                n_samples,
                n_marginal_parameters,
            ), "size of the marginal parameters does not match"

    @pytest.mark.parametrize(
        "n_samples, marginal_indices, batch_size",
        product(
            [5, 10],
            [[0], [0, 1], [(0, 1)]],
            [None, 10],
        ),
    )
    def test_sample_shape(
        self, n_samples: int, marginal_indices: MarginalIndex, batch_size: Optional[int]
    ):
        np.random.seed(0)
        torch.manual_seed(0)
        marginal_indices = tupleize_marginal_indices(marginal_indices)
        n_marginal_parameters = len(marginal_indices[0])
        marginal_ratio_estimator = self.get_marginal_ratio_estimator(marginal_indices)

        prior = get_diagonal_normal_prior(
            loc=[0] * self.n_parameters, scale=[1] * self.n_parameters
        )

        marginal_posterior = MarginalPosterior(marginal_ratio_estimator, prior)

        fabricated_observation = {
            key: torch.rand(*shape) for key, shape in self.observation_shapes.items()
        }
        samples = marginal_posterior.sample(
            n_samples=n_samples,
            observation=fabricated_observation,
            batch_size=batch_size,
        )
        assert set(samples.keys()) == set(marginal_indices)
        for _, value in samples.items():
            assert value.shape == (n_samples, n_marginal_parameters)


if __name__ == "__main__":
    pass



from swyft.networks.channelized import (
    BatchNorm1dWithChannel,
    LinearWithChannel,
    ResidualNetWithChannel,
)
from swyft.networks.classifier import (
    MarginalClassifier,
    Network,
    ObservationTransform,
    ParameterTransform,
    get_marginal_classifier,
)
from swyft.networks.standardization import (
    OnlineDictStandardizingLayer,
    OnlineStandardizingLayer,
)

__all__ = [
    "BatchNorm1dWithChannel",
    "LinearWithChannel",
    "ResidualNetWithChannel",
    "OnlineDictStandardizingLayer",
    "OnlineStandardizingLayer",
    "ObservationTransform",
    "ParameterTransform",
    "MarginalClassifier",
    "Network",
    "get_marginal_classifier",
]



#!/usr/bin/env python3

import sys

import simulators
import torch


def main():
    try:
        infile = sys.argv[1]
        outfile = sys.argv[2]
    except:
        print("Usage: simulate.py infile outfile")
    model = simulators.model_FermiV1
    params = torch.load(infile)
    obs = model(params)
    torch.save(obs, outfile)


if __name__ == "__main__":
    main()



from typing import Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from scipy import stats

from swyft.plot.histogram import split_corner_axes
from swyft.types import Array, MarginalToArray
from swyft.utils.marginals import get_d_dim_marginal_indices


def get_z_score(alpha: Union[float, np.ndarray]) -> np.ndarray:
    """Recover the z_score given by `z = normal_ppd(1 - alpha / 2)`.

    Args:
        alpha: significance level

    Returns:
        z_score
    """
    return stats.norm.ppf(1 - alpha / 2)


def get_alpha(z_score: Union[float, np.ndarray]) -> np.ndarray:
    """Recover the alpha (significance level) given by `alpha = 2 * (1 - normal_cdf(z_score))`.

    Args:
        z_score: z_score aka `z`

    Returns:
        alpha: significance level
    """
    return 2 * (1 - stats.norm.cdf(z_score))


def get_jefferys_interval(
    n_success: Union[int, np.ndarray],
    n_trials: int,
    alpha: Union[float, np.ndarray] = 0.05,
) -> np.ndarray:
    """Estimate the lower and upper jefferys confidence intervals.

    Args:
        n_success: number of successes. shape=(n,)
        n_trials: number of trials
        alpha: significance level. Defaults to 0.05. shape=(n,)

    Returns:
        shape=(n, 2), lower confidence interval, upper confidence interval

    Citation:
        Brown, Lawrence D., T. Tony Cai, and Anirban DasGupta. "Interval estimation for a binomial proportion." Statistical science 16.2 (2001): 101-133.
    """
    lower = stats.beta.ppf(alpha / 2, n_success + 0.5, n_trials - n_success + 0.5)
    upper = stats.beta.ppf(1 - alpha / 2, n_success + 0.5, n_trials - n_success + 0.5)
    return np.stack(
        [
            np.where(n_success > 0, lower, 0.0),
            np.where(n_success < n_trials, upper, 1.0),
        ],
        axis=-1,
    )


def get_empirical_z_score(
    empirical_mass: Array,
    max_z_score: float,
    n_bins: int = 50,
    interval_z_score: float = 1.0,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute the empirical z-score and corresponding Jeffery's interval given mass estimates.

    Args:
        empirical_mass: empirical mass containing the parameter
        max_z_score: max limit to compute nominal z-score
        n_bins: number of nominal z-scores to compute. Defaults to 50.
        band_z_score: z-score of Jeffery's interval to plot. 1 implies one standard deviation of normal. Defaults to 1.0.

    Returns:
        nominal z-scores, mean empirical z-score, interval empirical z-score
    """
    empirical_mass = np.asarray(empirical_mass)
    n, *_ = empirical_mass.shape

    # create x axis z scores / confidences
    nominal_z_scores = np.linspace(0.0, max_z_score, n_bins)
    confidences = 1 - get_alpha(nominal_z_scores)

    # determine what counts as a success for us
    n_not_containing_truth = np.sum(empirical_mass[:, None] < confidences, axis=0)

    # compute the properties of the mean and Jeffery's interval
    mean = n_not_containing_truth / n
    interval = get_jefferys_interval(
        n_not_containing_truth, n, alpha=get_alpha(interval_z_score)
    )
    z_mean = get_z_score(1 - mean)
    z_interval = get_z_score(1 - interval)
    return nominal_z_scores, z_mean, z_interval


def plot_empirical_z_score(
    axes: Axes,
    nominal_z_scores: np.ndarray,
    z_mean: np.ndarray,
    z_interval: np.ndarray,
    mean_color: str = "black",
    interval_color: str = "0.8",
    diagonal_color: str = "darkgreen",
    sigma_color: str = "red",
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    xlabel: Optional[str] = "Nominal coverage [z-score]",
    ylabel: Optional[str] = "Empirical coverage [z-score]",
    diagonal_text: bool = False,
) -> Axes:
    """target a particular matplotlib Axes and produce an empirical coverage test plot with Jeffrey's interval

    Args:
        axes: matplotlib axes
        nominal_z_scores: sorted array of nominal z-scores
        z_mean: empirical mean of z-score estimate using a binominal distribution
        z_interval: jeffrey's interval of z-score estimate
        mean_color: color of the mean line.
        interval_color: color of the interval, floats are grey.
        diagonal_color: color of the diagonal, nominal z-score.
        sigma_color: color of the vertical and horizontal sigma lines.
        xlim: force xlim
        ylim: force ylim
        xlabel: set xlabel
        ylabel: set ylabel
        diagonal_text: turns on semantic description of above / below diagonal

    Returns:
        the matplotlib axes given
    """
    lower = z_interval[:, 0]
    upper = z_interval[:, 1]
    assert np.all(lower <= upper), "the lower interval must be <= the upper interval."
    upper = np.where(upper == np.inf, 100.0, upper)

    # empirical lines & interval
    axes.plot(nominal_z_scores, z_mean, color=mean_color)
    axes.fill_between(nominal_z_scores, lower, upper, color=interval_color)

    # diagonal line
    max_z_score = np.max(nominal_z_scores)
    axes.plot([0, max_z_score], [0, max_z_score], "--", color=diagonal_color)

    # horizontal and vertical lines, vertical are the "truth", horizontal are empirical
    for i_sigma in range(1, int(max_z_score) + 1):
        empirical_i_sigma = np.interp(i_sigma, nominal_z_scores, z_mean)
        if empirical_i_sigma != np.inf:  # when the vertical line intersects z_mean
            # Horizontal line
            axes.plot(
                [0, i_sigma],
                [empirical_i_sigma, empirical_i_sigma],
                ":",
                color=sigma_color,
            )
            # horizontal text
            c = 1 - get_alpha(empirical_i_sigma)
            axes.text(0.1, empirical_i_sigma + 0.05, ("%.2f" % (c * 100)) + "%")
            # vertical line
            axes.plot(
                [i_sigma, i_sigma], [0, empirical_i_sigma], ":", color=sigma_color
            )
            # vertical text
            c = 1 - get_alpha(i_sigma)
            axes.text(i_sigma, 0.3, ("%.2f" % (c * 100)) + "%", rotation=-90)
        else:  # when the vertical line fails to intersect z_mean
            # # horizontal line
            # axes.plot([i_sigma, i_sigma], [0, 10.0], ":", color=sigma_color)
            # # horizontal text
            # c = 1 - get_alpha(i_sigma)
            # axes.text(i_sigma, 0.3, ("%.2f" % (c * 100)) + "%", rotation=-90)
            pass

    # set labels
    axes.set_ylabel(xlabel)
    axes.set_xlabel(ylabel)

    # Add the semantic meaning of being above / below diagonal
    if diagonal_text:
        raise NotImplementedError("must add rotation description")
        phi = None
        axes.text(
            max_z_score / 2,
            max_z_score / 2 + 0.4,
            "Conservative",
            ha="center",
            va="center",
            rotation=phi,
            color="darkgreen",
        )
        axes.text(
            max_z_score / 2 + 0.4,
            max_z_score / 2,
            "Overconfident",
            ha="center",
            va="center",
            rotation=phi,
            color="darkgreen",
        )

    # set limits
    if xlim is None:
        axes.set_xlim([0, max_z_score])
    else:
        axes.set_xlim(xlim)

    if ylim is None:
        axes.set_ylim([0, max_z_score + np.round(0.15 * max_z_score, 1)])
    else:
        axes.set_ylim(ylim)
    return axes


def empirical_z_score_corner(
    empirical_mass_1d: MarginalToArray,
    empirical_mass_2d: MarginalToArray,
    max_z_score: float = 3.5,
    labels: Sequence[str] = None,
    figsize: Optional[Tuple[float, float]] = None,
    space_between_axes: float = 0.1,
) -> Tuple[Figure, Axes]:
    """create a corner plot with each subplot containing a p-p test of expected coverage probability

    Args:
        empirical_mass_1d: empirical mass dict from marginal_indices to array
        empirical_mass_2d: empirical mass dict from marginal_indices to array
        max_z_score: maximum nominal z-score to plot. Defaults to 3.5.
        labels: for labeling the x and y axes. Defaults to None.
        figsize: set figsize like in `plt.subplots`. Defaults to None.
        space_between_axes: changes the `wspace` and `hspace` between subplots. see `plt.subplots_adjust`. Defaults to 0.1.

    Returns:
        matplotlib figure, np array of matplotlib axes
    """
    # compute the dimension, necessary inds, assert all is there
    d = len(empirical_mass_1d)
    upper_inds = get_d_dim_marginal_indices(d, 2)
    assert len(empirical_mass_2d) == len(upper_inds)

    # create the plots, adjust them, remove unnecessary upper corner
    fig, axes = plt.subplots(
        nrows=d, ncols=d, sharex="col", sharey="row", figsize=figsize
    )

    _, diag, upper = split_corner_axes(axes)
    lb = 0.125
    tr = 0.9
    fig.subplots_adjust(
        left=lb,
        bottom=lb,
        right=tr,
        top=tr,
        wspace=space_between_axes,
        hspace=space_between_axes,
    )

    for ax in upper:
        ax.axis("off")

    # plot the 1d mass
    for i, (k, ax) in enumerate(zip(empirical_mass_1d.keys(), diag)):
        nominal_z_scores, z_mean, z_interval = get_empirical_z_score(
            empirical_mass_1d[k], max_z_score
        )
        plot_empirical_z_score(
            ax,
            nominal_z_scores,
            z_mean,
            z_interval,
            xlabel=None,
            ylabel=None,
        )

    # plot the 2d mass
    # for i, (k, ax) in enumerate(zip(empirical_mass_2d.keys(), upper)):
    for k, i in zip(empirical_mass_2d.keys(), upper_inds):
        a, b = i  # plot array index, upper index (lower index is transpose)
        ax = axes[b, a]  # targets the lower left corner

        nominal_z_scores, z_mean, z_interval = get_empirical_z_score(
            empirical_mass_2d[k], max_z_score
        )
        plot_empirical_z_score(
            ax,
            nominal_z_scores,
            z_mean,
            z_interval,
            xlabel=None,
            ylabel=None,
        )

    # bottom row
    for i, ax in enumerate(axes[-1, :]):
        if labels is not None:
            ax.set_xlabel(labels[i])

    # left column
    for i, ax in enumerate(axes[1:, 0], 1):
        if labels is not None:
            ax.set_ylabel(labels[i])

    fig.align_labels()
    return fig, axes


if __name__ == "__main__":
    pass
